/***********************************************************************

AGB-DTVD v0.04 BETA by Herman Haverkort, 5 April 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-DTVD may be available from 
http://herman.haverkort.net.


************************************************************************

Was ist AGB-DTVD?

AGB-DTVD implements two algorithms to compute Voronoi diagrams and 
Delaunay triangulations. Its purpose is to illustrate the algorithms
described in Chapter 6 of Klein, Driemel and Haverkort: Algorithmische 
Geometrie, 3rd edition, 2022. AGB-DTVD consists of six files:
- agb-dtvd-main.cpp
- agb-dtvd-io.cpp (this file: pre- and postprocessing operations)
- agb-dtvd-dc.cpp (the Divide-and-Conquer algorithm)
- agb-dtvd-ricdd.cpp (randomised incremental construction with the 
    Delaunay DAG)
- agb-simplepoint.cpp (the basis geometric operations)
- agb-qeds.cpp (the quad-edge data structure)


***********************************************************************/

#ifndef AGB_DTVD_INCLUDED
#define AGB_DTVD_INCLUDED

/**********************************************************************/

#include "agb-qeds.cpp"
using AGBQEDS::HalfEdge;
#include "agb-simplepoint.cpp"
#include <algorithm>
using std::sort;


/**********************************************************************/

// VORONOI DIAGRAM AND DELAUNAY TRIANGULATION

// In the Voronoi diagram, there are:
// - vertices with floating point coordinates; these will only be 
//   calculated at the end, when preparing postscript output, not in 
//   the graph construction phase;  
// - faces that correspond to sites with exact coordinates.
// In the Delaunay triangulation, the roles of vertices and faces are
// the other way around.

// For the Voronoi regions/sites, we use the SPoint class from 
// agb-simplepoint.cpp:

typedef AGB::SPoint Site;


// Intialise a point at infinity (or: on the back)

Site infinity = Site::pointAtInfinity();


// The class that we will use for the Voronoi vertices / Delaunay 
// triangles, will depend on the algorithm (the randomised incremental
// construction algorithm need to link the triangles with the Delaunay
// DAG) and will be defined in the corresponding source file.


/**********************************************************************/

// SORTING INPUT POINTS

// A class that implements a lexicographic comparison operator:
struct LexicographicOrder {
  bool operator()(const Site& a, const Site& b) const { 
    if (a.getX() < b.getX()) return true;
    if (a.getX() > b.getX()) return false;
    return (a.getY() < b.getY());
  }
};

// Sort lexicographically and remove duplicates:
void sortAndRemoveDuplicates(std::vector<Site>& sites)
{
  sort(sites.begin(), sites.end(), LexicographicOrder());
  std::vector<Site> uniqueSites;
  uniqueSites.push_back(sites.front());
  for (
    std::vector<Site>::const_iterator i = sites.begin();
    i != sites.end();
    ++i
  )
  {
    // only keep sites that differ from the previous one:
    if (uniqueSites.back() != *i) uniqueSites.push_back(*i);
  }
 
  // replace the sorted input by the sorted input without
  // duplicates:
  sites = uniqueSites;
}


/**********************************************************************/

// RANDOM PERMUTATION

// the permutation function requires a pointer to an array of sites 
// and the number of sites to permute:
void permuteRandomly(Site sites[], int n)
{
  // go through all elements from the back to the front and swap each 
  // element with a random element out of those not considered yet; 
  // this produces each possible permutation with probability 1/n!

  for (int i = n; i > 1; --i)
    swap(sites[i-1], sites[rand() % i]);
}


/**********************************************************************/

// COLOURING THE VORONOI REGIONS

// The following function calculates a 6-colouring of the sites such 
// that adjacent sites have different colours. It is possible to do 
// with fewer colours. In fact, it seems that simply changing 5 and 6 
// by 4 and 5 in the code below usually does the trick already---but to
// get any guarantees, we need to a bit more coding.

// The function is templated because it needs to work for diagrams with
// different classes for the Voronoi vertices / Delaunay triangles 
// (with or without a Delaunay DAG).

template < class VoronoiVertex >
void colourSites(HalfEdge<VoronoiVertex,Site> voronoiDiagram)
{
  // actually it is easier to work with the dual graph, that is, the
  // Delaunay triangulation. We get a list of halfedges in the dual, 
  // for each site we get one outgoing halfedge to represent it:
  typedef HalfEdge<Site,VoronoiVertex> DelaunayEdge;
  std::vector<DelaunayEdge> regions = 
    voronoiDiagram.rtol().getVertices();

  // every simple planar graph has at least one vertex of degree 
  // at most five, so we could deconstruct the graph by removing 
  // vertices (sites) of degree at most five, until no sites are 
  // left. We are not actually going to take the graph apart, but 
  // we will build an order in which we could remove the sites
  // on the stack colouringOrder;
  std::stack<DelaunayEdge> colouringOrder;

  // sites eligible to be put on colouringOrder are those who
  // have at most five neighbours that are not on colouringOrder
  // yet; we will keep these eligible sites in toBeScheduled:
  std::stack<DelaunayEdge> toBeScheduled;

  // first count for each site, how many neighbours it has;
  // we will abuse the colour attribute to store the number.
  // Those that have at most five neighbours can be pushed onto
  // toBeScheduled right away; we will set their colour attribute 
  // to zero:
  for (
    typename std::vector<DelaunayEdge>::iterator i = regions.begin();
    i != regions.end();
    ++i
  )
  {
    i->origin()->colour = i->nrEdgesOnOrigin();
    if (i->origin()->colour > 5) continue;
    i->origin()->colour = 0;
    toBeScheduled.push(*i);
  }

  // now we keep moving sites from toBeScheduled to colouringOrder;
  // whenever we move a sites, we lower the colour (the neighbours-not-
  // on-colouringOrder counter) for all neighbours that are not on 
  // toBeScheduled or colouringOrder themselves yet; any neighbours 
  // whose count is reduced to five are pushed onto toBeScheduled:
  while (!toBeScheduled.empty())
  {
    DelaunayEdge e = toBeScheduled.top(); toBeScheduled.pop();
    colouringOrder.push(e);

    // initialise an edge f which we are going to use to iterate over
    // all edges with the same origin as e; we will treat e itself first:
    DelaunayEdge f = e;
    do
    {
      if (f.destin()->colour == 0) 
        // neighbour is on toBeScheduled or colouringOrder already,
        // nothing needs to be done for this neighbour:
        continue;

      // update the neighbour's counter: we just moved the origin of e
      // from toBeScheduled to colouringOrder, so for the neighbour,
      // the number of neighbours not on colouringOrder decreased by 1:

      --f.destin()->colour;

      if (f.destin()->colour > 5) 
        // neighbour cannot be moved to toBeScheduled yet:
        continue;

      // now we have: f.destin()->colour == 5. Push this neighbour onto 
      // toBeScheduled; set colour to 0 to record this:

      toBeScheduled.push(f.rev());
      f.destin()->colour = 0;

    } while (++f != e); // proceed to the next edge, until back at e.
  }

  // now we can assign colours to sites (= regions) in the reverse order 
  // in which they were pushed onto colouringOrder; whenever we take a 
  // site from the stack, we know it has at most five neighbours that 
  // were pushed on the stack later and have therefore been coloured 
  // already. This implies that at least one out of six available 
  // colours has not been used for one of the neighbours yet, so we can 
  // choose it for the current site.

  // we are going to keep track of how many sites have which colour 
  // already, so that, in case we have a choice, we can choose the colour
  // that has been used least so far (this is just a heuristic to get a
  // balanced use of colour):

  int nrWithColour[7]; // where ...[0] is the total number of sites
  nrWithColour[0] = regions.size();
  for (int i = 1; i <= 6; ++i) nrWithColour[i] = 0;

  // now colour all sites from top to bottom on the stack:
  while (!colouringOrder.empty())
  {
    DelaunayEdge e = colouringOrder.top(); colouringOrder.pop();

    bool available[7];
    for (int i = 1; i <= 6; ++i) available[i] = true;

    // check all neighbours and mark the colours used by them as 
    // unavailable:

    DelaunayEdge f = e;
    do available[f.destin()->colour] = false; while (++f != e);

    // now pick a colour; the first available colour will always be 
    // taken (since initially, e.origin()->colour == 0 and 
    // nrWithColour[i] is always smaller than nrWithColour[0]).
    // After that, we might replace it by another colour if it has
    // been used less often yet:

    for (int i = 1; i <= 6; ++i)
      if (available[i] && nrWithColour[i] < nrWithColour[e.origin()->colour])
        e.origin()->colour = i;

    // update the count:
    ++nrWithColour[e.origin()->colour];
  }
}


/**********************************************************************/

// POINTS FOR GRAPHICS OUTPUT

// For graphics output there is a point class that store coordinates as
// floating point numbers; these do not support exact computations.

struct FPoint 
{
  double x;
  double y;
  
  // constructors:
  FPoint(double _x = 0, double _y = 0): 
    x(_x), y(_y) 
  {};
  FPoint(const AGB::SPoint& p): 
    x(p.getX()), y(p.getY()) 
  {};

  // various operators
  FPoint operator+(const FPoint& p) const {
    return FPoint(x+p.x, y+p.y); 
  }
  FPoint operator-(const FPoint& p) const {
    return FPoint(x-p.x, y-p.y); 
  }
  FPoint operator*(double s) const {
    return FPoint(x*s, y*s);
  } 
  double square() const {
    return x*x + y*y;
  }
};


// operator to write the point coordinates to a (postscript) file:

ostream& operator<<(ostream& os, const FPoint& p)
{
  return os << p.x << ' ' << p.y;
}


/**********************************************************************/

// COMPUTING THE COORDINATES OF THE VORONOI VERTICES

// the following function computes a point on the bisector of u and
// v, to the left of the directed line u, v and far enough away to be 
// outside the viewing window:

FPoint farAwayToTheLeftOnBisector(
  const AGB::SPoint& u, const AGB::SPoint& v)
{
  assert(u != v);
  assert(!u.atInfinity());
  assert(!v.atInfinity());
  FPoint midPoint = (FPoint(u)+FPoint(v))*0.5;
  FPoint direction = FPoint(v)-FPoint(u);
  double scale = (double) Site::INF / sqrt(direction.square());
  return FPoint(
    midPoint.x - direction.y * scale, 
    midPoint.y + direction.x * scale
  );
}


// the following function computes the centre of the circumscribed
// circle of three given points, given in counterclockwise order;
// we will need it to compute the coordinates of Voronoi vertices

FPoint centrePoint(
  const AGB::SPoint& u, 
  const AGB::SPoint& v, 
  const AGB::SPoint& w
)
{
  // The centre point is the solution of a system of linear equation:
  // u and v determine a bisector and v and w determine a bisector;
  // the centre point is their intersection point. I rewrote the 
  // solution such that for each coordinate, we need only one 
  // division:
  double d = 2 * doubleSignedArea(u, v, w);
  assert(d != 0); // verify that u, v, w are not collinear
  FPoint uv = FPoint(v) - FPoint(u);
  FPoint vw = FPoint(w) - FPoint(v);
  FPoint wu = FPoint(u) - FPoint(w);
  return FPoint(
    (-vw.y * u.square() - wu.y * v.square() - uv.y * w.square())/d,
    ( vw.x * u.square() + wu.x * v.square() + uv.x * w.square())/d
  );
}


// The following function computes all vertex coordinates.
// The function is templated because it needs to work for different
// types of diagrams (with or without a Delaunay DAG).
// The Voronoi vertex type V must allow being assigned an FPoint, 
// that is, V& operator=(V& v, const FPoint& p) must be defined.

template < class V >
void computeVertexCoordinates(HalfEdge<V,Site> voronoiDiagram)
{
  // Get a list of all vertices, that is, for each vertex an half-edge
  // originating from it:
  std::vector<HalfEdge<V,Site> > vertices = voronoiDiagram.getVertices();

  // Compute all vertex coordinates:
  for (
    typename std::vector<HalfEdge<V,Site> >::iterator i = vertices.begin();
    i != vertices.end();
    ++i
  )
  {
    // Get three sites whose regions meet at the origin of *i: 
    Site& l = *i->leftFace();  
    Site& r = *i->rightFace(); 
    Site& o = *i->sccorgn().leftFace();

    // make an object to store the vertex info, if it did not exist yet:
    if (i->origin() == NULL) i->labelOrigin(new V);

    // if one of the sites lies at infinity, we compute a point on 
    // the bisector of the other two sites that lies far enough
    // away for our purposes, otherwise we just compute the centre of
    // the circumcircle of the three sites:

    if (l.atInfinity()) 
      *i->origin() = farAwayToTheLeftOnBisector(o, r); 
    else if (o.atInfinity()) 
      *i->origin() = farAwayToTheLeftOnBisector(r, l); 
    else if (r.atInfinity()) 
      *i->origin() = farAwayToTheLeftOnBisector(l, o); 
    else 
      *i->origin() = centrePoint(l,o,r);
  }
}


/**********************************************************************/

// POSTSCRIPT OUTPUT

// The following function takes an edge from a Voronoi diagram and 
// writes a square postscript drawing, measuring 500 x 500 DTP points, 
// showing the complete diagram, to the specified output stream. It also 
// needs the coordinates of the lower left and the upper right bounding 
// box of the sites---these are used to scale the drawing such that the 
// sites lie well within its boundaries.
 
// The function is templated because it needs to work for diagrams with
// different classes for the Voronoi vertices / Delaunay triangles 
// (with or without a Delaunay DAG).

template < class V >
void drawVoronoiDiagram(
  HalfEdge<V,Site> vd, const FPoint& min, const FPoint& max,
  ostream& os
)
{
  // the following object provides:
  // - an operator that scales and translates the sites so that they
  //   all lie within the window [50,450]x[50,450]
  // - a function compassDirection that determines, for a given 
  //   point, in which quadrant it lies w.r.t. the diagonals of the 
  //   drawing: 0 = east, 1 = north, 2 = west, or 3 = south.

  struct Map {
    double scale;
    FPoint centre;
    Map(const FPoint& min, const FPoint& max):
      centre((min + max)*0.5),
      scale(std::min(
        400.0 / (max.x-min.x+1),
        400.0 / (max.y-min.y+1)
      ))
    {};
    FPoint operator()(const FPoint& p) const {
      return (p - centre)*scale + FPoint(250,250);
    }
    int compassDirection(const FPoint& p) const {
      FPoint rp = p - centre;
      return (rp.y > rp.x ? 1 : 0) ^ (rp.y > -rp.x ? 0 : 3);
    }
  } map = Map(min, max);

  // these are the RGB values of the colours:
  const char* colour[7] = {
    "1.00 1.00 1.00", // white, just in case, not really to be used
    "0.90 0.10 0.30", // red
    "0.25 0.70 0.30", // green
    "1.00 1.00 0.10", // yellow
    "0.00 0.50 0.80", // blue
    "0.95 0.50 0.20", // orange
    "0.55 0.10 0.70"  // magenta
  };

  // these are points just outside the corners of the viewing window
  // [0,500] x [0,500], we will need them to draw the regions that
  // cover the corners correctly:
  const FPoint corner[4] = {
    FPoint( 600,  600),
    FPoint(-100,  600),
    FPoint(-100, -100),
    FPoint( 600, -100)
  };

  // write the initial part of the postscript file:
  os << "\%!PS-Adobe-3.0 EPSF-3.0" << endl
     << "\%\%BoundingBox: 0 0 500 500" << endl
     << "\%\%HiResBoundingBox: 0 0 500 500" << endl
     << "gsave" << endl
     << "newpath 0 0 moveto 0 500 lineto 500 500 lineto 500 0 lineto 0 0 lineto clip" 
     << endl
     << "1 setlinejoin" << endl
     << "0.2 setlinewidth" << endl;

  // collect a list of edges from the dual of the Voronoi diagram, 
  // for each site one outgoing edge:
  std::vector<HalfEdge<Site,V> > regions = vd.rtol().getVertices();

  // draw the Voronoi regions:
  for (
    typename std::vector<HalfEdge<Site,V> >::iterator i = regions.begin(); 
    i != regions.end(); 
    ++i
  ) 
  {
    if (i->origin()->atInfinity()) continue;
    os << colour[i->origin()->colour] << " setrgbcolor "
       << "newpath " << map(*i->rtol().origin()) << " moveto ";

    // iterate over all corners of the region; for this purpose we 
    // iterate over all dual edges that originate from the site; the 
    // faces to the left and right of these dual edges correspond to 
    // the vertices of the Voronoi region's polygon:
    HalfEdge<Site,V> e = *i;
    do 
    {
      // the tricky part is that we have to make sure that edges
      // at infinity, connecting far-away points on two bisectors, 
      // do not cut off one or more corners of the drawing window:
      if (e.destin()->atInfinity())
      {
        int fromCompassDirection = map.compassDirection(*e.rightFace());
        int toCompassDirection = map.compassDirection(*e.leftFace());
        while (fromCompassDirection != toCompassDirection)
        {
          // make sure to go around the corner:
          os << corner[fromCompassDirection] << " lineto ";
          ++fromCompassDirection; fromCompassDirection %= 4;
        }
      }
      // after rounding any corners, proceed to the next vertex:
      os << map(*e.rtol().destin()) << " lineto ";
    } 
    while (++e != *i);

    // finish the region's polygon:
    os << "gsave fill grestore 1 setgray stroke" << endl;
  }

  // draw the Delaunay triangulation:
  std::cout << "0.2 setlinewidth 0 setgray" << endl;
  std::vector<HalfEdge<Site, V> > edges = vd.rtol().getEdges();
  for (
    typename std::vector<HalfEdge<Site, V> >::iterator i = edges.begin(); 
    i != edges.end(); 
    ++i
  ) 
  {
    if (i->origin()->atInfinity()) continue;
    if (i->destin()->atInfinity()) continue;
    os << "newpath " << map(FPoint(*i->origin())) << " moveto " 
       << map(FPoint(*i->destin())) << " lineto stroke" << endl;
  }

  // draw the sites:
  for (
    typename std::vector<HalfEdge<Site,V> >::iterator i = regions.begin(); 
    i != regions.end(); 
    ++i
  ) 
  {
    if (i->origin()->atInfinity()) continue;
    os << colour[i->origin()->colour] << " setrgbcolor ";
    os << map(FPoint(*i->origin())) << " 0.1 0 360 arc closepath fill" << endl;  
  }

  // finish the postscript file:
  os << "grestore" << endl;
}


/**********************************************************************/

#endif

