/***********************************************************************

AGB-DTVD v0.04 BETA by Herman Haverkort, 5 April 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-DTVD may be available from 
http://herman.haverkort.net.


************************************************************************

Was ist AGB-DTVD?

AGB-DTVD implements two algorithms to compute Voronoi diagrams and 
Delaunay triangulations. Its purpose is to illustrate the algorithms
described in Chapter 6 of Klein, Driemel and Haverkort: Algorithmische 
Geometrie, 3rd edition, 2022. AGB-DTVD consists of six files:
- agb-dtvd-main.cpp (this file)
- agb-dtvd-io.cpp (pre- and postprocessing operations)
- agb-dtvd-dc.cpp (the Divide-and-Conquer algorithm)
- agb-dtvd-ricdd.cpp (randomised incremental construction with the 
    Delaunay DAG)
- agb-simplepoint.cpp (the basis geometric operations)
- agb-qeds.cpp (the quad-edge data structure)


***********************************************************************/

#include <iostream>
using std::cerr;
using std::endl;
#include <ctime>

#include "agb-dtvd-io.cpp"

#ifdef INCLUDE_AGB_DTVD_DC
#include "agb-dtvd-dc.cpp"
#else
#ifdef INCLUDE_AGB_DTVD_RICDD
#include "agb-dtvd-ricdd.cpp"
#else
#error No Delaunay triangulation or Voronoi diagram algorithm specified. Please compile with -D INCLUDE_AGB_DTVD_DC or with -D INCLUDE_AGB_DTVD_RICDD.
#endif
#endif


/**********************************************************************/

// a variable and two functions to keep track of the time:

clock_t started;

void start()
{
  cerr << "..."; cerr.flush();
  started = clock();
}

void finish()
{
  cerr << " done in " 
    << 1000.0*(clock()-started)/CLOCKS_PER_SEC << "ms" << endl;
}


/**********************************************************************/

// a function to keep track of the lower left and the upper right
// corner of the bounding box of the points read so far:

void updateMinMax(const FPoint& p, FPoint& min, FPoint& max)
{
  if (p.x < min.x) min.x = p.x;
  if (p.y < min.y) min.y = p.y;
  if (p.x > max.x) max.x = p.x;
  if (p.y > max.y) max.y = p.y;
}


/**********************************************************************/

int main() 
{
  // read the input and record the lower left corner and the upper right
  // corner of the bounding box in min, max:

  int nrPoints; std::cin >> nrPoints;
  cerr << "Reading " << nrPoints << " points; "
    << "coordinates must lie in [" 
    << -Site::W << ',' << Site::W << "]";
  start();
  std::vector<Site> points(nrPoints);
  FPoint min( Site::INF,  Site::INF);
  FPoint max(-Site::INF, -Site::INF);
  for (int i = 0; i < nrPoints; ++i) 
  {
    std::cin >> points[i];
    updateMinMax(FPoint(points[i]), min, max);
  }
  finish();

  // process the input:

  cerr << "Sorting and removing duplicates"; 
  start();
  sortAndRemoveDuplicates(points);
  finish();
  cerr << points.size() << " distinct points in [" 
       << min.x << ',' << max.x << "] x [" << min.y << ',' << max.y
       << "] remain." << endl;

  cerr << "Computing Voronoi diagram and Delaunay triangulation by "
       << dtvdAlgorithmDescription;
  start(); 
  // computeVD computes the graph structure of the Voronoi diagram 
  // and the Delaunay triangulation, and returns a half-edge of the
  // Voronoi diagram as a representative:
  VoronoiEdge vd = computeVD(points);
  finish();

  cerr << "Computing vertex coordinates"; 
  start();
  computeVertexCoordinates(vd);
  finish();

  cerr << "Computing a colouring of the regions"; 
  start();
  colourSites(vd);
  finish();

  cerr << "Computing postscript output"; 
  start();
  // use the corners of the bounding box of the input to determine
  // a reasonable clipping window:
  drawVoronoiDiagram(vd, min, max, std::cout);
  finish();

  // The final step is just to make sure that we did not create any 
  // memory leaks in the form of the inaccessible QEDS fragments.
  // Since we are at the end of the program, this step does not 
  // actually do anything useful; the memory will be released anyway.
  cerr << "Cleaning up"; 
  start();
  VoronoiEdge::deleteGraph(vd);
  finish();
  if (AGBQEDS::Edge::totalNrAlive() != 0) // this should not happen!
    cerr << " OOPS: " << AGBQEDS::Edge::totalNrAlive() 
      << " QEDS edges remain and could not be found!?"
      << endl;
}
