/***********************************************************************

AGB-SimplePoint v0.02 BETA by Herman Haverkort, 2 April 2022

************************************************************************

Copyright 2022 Herman Johannes Haverkort

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

New versions of AGB-SimplePoint may be available from 
http://herman.haverkort.net.


************************************************************************

Was ist AGB-SimplePoint?

AGB-SimplePoint implements basic geometric operations on points that are
useful in computing Voronoi diagrams and Delaunay triangulations.
AGB-SimplePoint is meant as proof-of-concept software for educational 
purposes; it supports a limited coordinate range for input points.


***********************************************************************/

#ifndef AGB_SIMPLEPOINT_INCLUDED
#define AGB_SIMPLEPOINT_INCLUDED

/**********************************************************************/

#include <limits>
#include <cassert>
#include <iostream>
using std::istream;
using std::ostream;
#include <cstdlib>


/**********************************************************************/

namespace AGB {


/**********************************************************************/

// A object of class SPoint stores a point in the plane with exact 
// coordinates, or a symbolic point at infinity. It also has an 
// attribute colour, which we can use any way we like.

class SPoint
{
  // In dieser Demonstration benutzen wir für unsere exakte Berechnungen
  // einfach ganze Zahlen:
    public:
  typedef long int Number;

  // Wenn die Koordinaten der Punkt in [-W,W] liegen, dann können die 
  // Zahlen in unseren Berechnungen in [-96 W^4,96 W^4] liegen 
  // (konservativ abgeschätzt), wie man in den Kommentaren zum Code 
  // nachvollziehen kann. Bei W = 17605 passt das in 64 Bits. Damit wir 
  // garantiert keine Overflow-Fehler bekommen, müssen wir darum die 
  // Koordinaten in der Eingabe entsprechend begrenzen. Dazu wird die
  // Konstante W benutzt:
  static const Number W;

  // We will use a specific larger number to denote infinity:
  static const Number INF;

  // The coordinates of the point are private: we do not want them to be
  // modified by external methods that may not guarantee that their 
  // values stay within [-W,W]. Access is through getX and getY:
    private:
  Number x; // in [-W, W]
  Number y; // in [-W, W]

    public:
  Number getX() const;
  Number getY() const;

  // The colour attribute is public (no need to keep it private):
    public:
  int colour;

    public:
  // A constructor to make a SPoint object with given coordinates:
  SPoint(Number _x = 0, Number _y = 0);

  // Points can also be read from an input stream (when given as
  // coordinates separated by a space):
  friend istream& operator>>(istream& i, SPoint& p);

  // A method to create a point at infinity:
  static SPoint pointAtInfinity();

  // Determine whether a point lies at infinity:
  bool atInfinity() const;

  // Comparison operators:
  bool operator==(const SPoint& p) const;
  bool operator!=(const SPoint& p) const;

  // the next operations are private because the result may violate 
  // the bounds on the number size; the resulting points can be ok
  // as temporaries in SPoint class operations, but they should 
  // not be used as normal SPoints.
    private:
  SPoint operator+(const SPoint& p) const;
  SPoint operator-(const SPoint& p) const;

    public:
  // inner product, in [-2*W^2, 2*W^2]
  Number operator*(const SPoint& p) const;

  // square norm (i.e. inner product with itself), in [0, 2*W^2]
  Number square() const; 

  // turned 90 degrees clockwise, that is, the point (y,-x):
  SPoint rightNormal() const;

  // decide whether the point lies in an open disk with u, v, w 
  // on its boundary in clockwise order, for any point w that lies 
  // far enough away; if u or v lies at infinity, returns true 
  // iff the *this point differs from the regular point out of 
  // u and v;
  bool inInfiniteDisk(const SPoint& u, const SPoint& v) const;

  // decide whether the point lies in an open disk with u, v, w
  // on the boundary in clockwise order.
  // behaviour undefined when u, v, w are collinear.
  bool inOpenDisk(
    const SPoint& u, const SPoint& v, const SPoint& w) const;

  // decide whether a point lies in a closed triangle with
  // vertices u, v, w in clockwise order.
  // behavour undefined when u, v, w are collinear.
  bool inClosedTriangle(
    const SPoint& u, const SPoint& v, const SPoint& w) const;
};

// determine the direction of turn of the sequence uvw:
bool leftTurn(const SPoint& u, const SPoint& v, const SPoint& w);
bool rightTurn(const SPoint& u, const SPoint& v, const SPoint& w);
bool collinear(const SPoint& u, const SPoint& v, const SPoint& w);
bool collinearInOrder(const SPoint& u, const SPoint& v, const SPoint& w);

// the following function calculates twice the area of the triangle 
// uvw (positive if in counterclockwise order; negative if in 
// clockwise order; zero otherwise); it used by several functions
// mentioned above:
long int doubleSignedArea(
  const SPoint& u, const SPoint& v, const SPoint& w);


/**********************************************************************/


const SPoint::Number SPoint::W = 
  pow(std::numeric_limits<Number>::max() / 96, 0.25);


const SPoint::Number SPoint::INF = 4 * W;


SPoint::Number SPoint::getX() const {
  return x;
}


SPoint::Number SPoint::getY() const {
  return y;
}


SPoint::SPoint(SPoint::Number _x, SPoint::Number _y): 
  x(_x), y(_y) 
{
  assert(-W <= x && x <= W);
  assert(-W <= y && y <= W);
}


// read a point from an input stream 
// (two coordinates separated by spaces):
istream& operator>>(istream& i, SPoint& p)
{
  i >> p.x >> p.y;

  // check the bounds:
  if (p.x < -SPoint::W || p.x > SPoint::W || p.y < -SPoint::W || p.y > SPoint::W)
  {
    std::cerr 
      << "Input point coordinates (" << p.x << ',' << p.y << ") out of range:" 
      << std::endl
      << "Must be in [" << -SPoint::W << ',' << SPoint::W 
      << "]x[" << -SPoint::W << ',' << SPoint::W << "]."
      << std::endl;
    exit(-1);
  }

  return i;
}


// create a point at infinity:
SPoint SPoint::pointAtInfinity()
{
  SPoint p;
  p.x = INF;
  p.y = INF;
  return p;
}


// decide whether a point lies at infinity:
bool SPoint::atInfinity() const
{
  return (x == INF);
}


bool SPoint::operator==(const SPoint& p) const 
{
  return (x == p.x && y == p.y);
}


bool SPoint::operator!=(const SPoint& p) const 
{
  return (x != p.x || y != p.y);
}


SPoint SPoint::operator+(const SPoint& p) const 
{
  assert(!atInfinity());
  assert(!p.atInfinity());
  SPoint r; 
  r.x = x + p.x; // in [-2*W, 2*W]
  r.y = y + p.y; // in [-2*W, 2*W]
  return r; 
}


SPoint SPoint::operator-(const SPoint& p) const 
{
  assert(!atInfinity());
  assert(!p.atInfinity());
  SPoint r; 
  r.x = x - p.x; // in [-2*W, 2*W]
  r.y = y - p.y; // in [-2*W, 2*W]
  return r; 
}


// inner product:
SPoint::Number SPoint::operator*(const SPoint& p) const
{
  assert(!atInfinity());
  assert(!p.atInfinity());
  return x * p.x + y * p.y; // in [-2*W^2, 2*W^2]
}


SPoint::Number SPoint::square() const 
{
  assert(!atInfinity());
  return x*x + y*y; // in [0, 2*W^2]
}


SPoint SPoint::rightNormal() const
{
  assert(!atInfinity());
  return SPoint(y,-x);
}


bool SPoint::inInfiniteDisk(const SPoint& u, const SPoint& v) const
{
  // not to be used on points at infinity:

  assert(!atInfinity());

  // if u or v lies at infinity, we consider *this to lie in the disk
  // whenever it differs from u and v:

  if (u.atInfinity()) return (*this != v);
  if (v.atInfinity()) return (*this != u);

  // otherwise *this lies in the disk when it lies to the left of 
  // the directed line through u and v, or when it lies on that line
  // between u and v: 

  return leftTurn(u, v, *this) || collinearInOrder(u, *this, v);
}


bool SPoint::inOpenDisk(
  const SPoint& u, const SPoint& v, const SPoint& w) const
{
  // not to be used on points at infinity:

  if (atInfinity()) return false;

  // if u, v, or w lies at infinity, then we test *this against the
  // "infinite disk" defined by the next two points in the counter-
  // clockwise order:

  if (u.atInfinity()) return inInfiniteDisk(v, w);
  if (v.atInfinity()) return inInfiniteDisk(w, u);
  if (w.atInfinity()) return inInfiniteDisk(u, v);

  // if all of u, v, and w are non-infinite points, then they should
  // not be collinear and they should be given in counterclockwise 
  // order, otherwise the result is undefined:

  assert(leftTurn(u, v, w));

  // to simplify the calculations, take all points relative to *this,
  // so that the question becomes: does the circumcircle of 
  // pu, pv, pw contain the origin?

  SPoint pu = u - *this; 
  SPoint pv = v - *this; 
  SPoint pw = w - *this;

  // we use the geometric transformation from Section 6.5 in the book:
  // for a point a = (a.x, a.y), let a' be (a.x, a.y, a.z), where
  // a.z = a.x*a.x + a.y*a.y. The origin lies in the circumcircle of 
  // pu, pv, pw, iff the plane through pu', pv', pw' passes above the
  // origin. So all we need to do is to compute the the z-coordinate 
  // at the origin by linear interpolation between pu', pv', and pw',
  // and check if it is more than zero.

  // for linear interpolation at a point O within a triangle pu pv pw 
  // (in counterclockwise order), we should take:

  return (

    // the value at pu, which is pu.square(), with weight proportional 
    // to the signed area of O pv pw; this weight is 
    // pv * pw.rightNormal() / 2 (see the doubleSignedArea function 
    // below, where we substitute O for u); we take twice that to 
    // avoid the division

    pu.square() * (pv * pw.rightNormal()) +

    // analogously, we take the z-coordinate of pv' with weight equal
    // to twice the signed area of O pw pu:

    pv.square() * (pw * pu.rightNormal()) +

    // and we take the z-coordinate of pw' with weight equal
    // to twice the signed area of O pu pv:

    pw.square() * (pu * pv.rightNormal())

    // thus we obtain the total weight (which is positive, as it is 
    // twice the area of the counterclockwise triangle pu pv pw) times 
    // the elevation of the plane through pu', pv', pw' above the 
    // origin. Since we only care whether the value is bigger than 
    // zero or not, we do not need to divide by the total weight, but
    // we simply compare to zero directly; if it is more than zero,
    // then the origin lies in the circumcircle of pu, pv, pw:

    > 0
  );

  // note that the sum computed above lies in [-96 W^4, 96 W^4]:
  // pu.square() is at most 8 W^2 (since |pu| is at most 2 W sqrt(2));
  // its weight lies in [-4 W^2, 4 W^2] (since it is twice the area of 
  // a triangle in the square [-W,W]x[-W,W];
  // so the first term lies in [-32 W^4, 32 W^4]; the same holds for
  // the second and the third term.
}


bool SPoint::inClosedTriangle(
    const SPoint& u, const SPoint& v, const SPoint& w) const
  {
  // not to be used on points at infinity:

  if (atInfinity()) return false;

  // if u, v, or w lies at infinity, then we test *this against the
  // "infinite disk" defined by the next two points in the counter-
  // clockwise order (one may also consider this to be a kind of 
  // triangle with one bounded edge and two infinite edges, that 
  // meet the bounded edge at angle almost pi and meet each other at
  // infinity):

  if (u.atInfinity()) return inInfiniteDisk(v, w);
  if (v.atInfinity()) return inInfiniteDisk(w, u);
  if (w.atInfinity()) return inInfiniteDisk(u, v);

  // if all of u, v, and w are non-infinite points, then they should
  // not be collinear and they should be given in counterclockwise 
  // order, otherwise the result is undefined:

  assert(leftTurn(u, v, w));

  // *this lies in the triangle iff it lies on or to the left of the
  // three lines that contain the edges of the triangle:

  return (
    !rightTurn(u, v, *this) &&
    !rightTurn(v, w, *this) &&
    !rightTurn(w, u, *this)
  );
}


// determine the direction of turn of the sequence uvw
bool leftTurn(const SPoint& u, const SPoint& v, const SPoint& w)
{
  return (doubleSignedArea(u, v, w) > 0);
}


bool rightTurn(const SPoint& u, const SPoint& v, const SPoint& w)
{
  return (doubleSignedArea(u, v, w) < 0);
}


bool collinear(const SPoint& u, const SPoint& v, const SPoint& w)
{
  return (doubleSignedArea(u, v, w) == 0);
}


bool collinearInOrder(const SPoint& u, const SPoint& v, const SPoint& w)
{
  if (!collinear(u, v, w)) return false;

  if (u == v || v == w) return false;
  // if u == w we should also return false, but the next line will
  // take care of that

  // if v lies between u and w, then the signs of vx-ux and vx-wx should
  // differ, and thus, (vx-ux)(vx-wx) < 0, unless u, v, w lie on a 
  // vertical line, in which case vx-ux = vx-wx = (vx-ux)(vx-wx) = 0.
  // the same goes for the y-coordinates;
  // given that we have already established that u, v, and w are collinear,
  // these conditions are not only necessary but also sufficient: 

  return (
    (v.getX()-u.getX())*(v.getX()-w.getX()) <= 0 && 
    (v.getY()-u.getY())*(v.getY()-w.getY()) <= 0
  );
}


// the following function calculates twice the area of the triangle 
// uvw (positive if in counterclockwise order; negative if in 
// clockwise order; zero otherwise); 
// the result lies in [-4 W^2, 4 W^2] because the input points lie 
// in the square [-W, W] x [-W, W]

long int doubleSignedArea(
  const SPoint& u, const SPoint& v, const SPoint& w) 
{
  // Imagine a base line that lies low enough at y = b, then the area 
  // above the baseline and vertically under an edge u, v that is 
  // directed to the left is (u.x-v.x)(u.y+v.y)/2; twice that is:
  // u.x u.y + u.x v.y - u.y v.x - v.x v.y
  // If the edge is directed from the left to the right, this 
  // expression gives the negative area. So if we add up these 
  // expressions for the edges (u,v) (v,w) (w,u), we add up positive
  // areas under the edges from the right to the left, in the upper 
  // part of the counterclockwise cycle, while substracting areas from
  // the left to the right, in the lower part of the counterclockwise
  // cycle; the net result is the double area of the triangle: 

  return u.getX() * v.getY() - u.getY() * v.getX()
       + v.getX() * w.getY() - v.getY() * w.getX()
       + w.getX() * u.getY() - w.getY() * u.getX();
}


/**********************************************************************/

} // closes namespace AGB

#endif
